#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "vm.h"
#include "image.h"

#define FILES_MAXIMUM 256

#define FILE_NAME_LENGTH 512

static FILE * files [FILES_MAXIMUM];

int
allocateFile (void)
{
        int file;

        for (file = 0; file < FILES_MAXIMUM; ++ file)
        {
                if (files [file] == NULL)
                  return file;
        }

        return -1;
}

void
closeFile (int file)
{
        if (files [file] != NULL)
        {
                fclose (files [file]);

                files [file] = NULL;
        }
}

int
openFile (struct ByteArray * name, char * mode)
{
        char nameString [FILE_NAME_LENGTH];
        size_t nameLength = PSObject_payloadSize ((struct Object *) name);
        int file;
        
        if (nameLength >= sizeof (nameString))
          return -1;

        memcpy (nameString, name -> elements, nameLength);
        nameString [nameLength] = '\0';

        file = allocateFile ();
        if (file < 0)
          return -1;

        files [file] = fopen (nameString, mode);
        if (files [file] == NULL)
          return -1;

        return file;
}

int
writeFile (int file, int n, char * bytes)
{
        return fwrite (bytes, 1, n, files [file]);
}

int
readFile (int file, int n, char * bytes)
{
        return fread (bytes, 1, n, files [file]);
}

int
sizeOfFile (int file)
{
       int pos = ftell (files [file]),
           size;
       fseek (files [file], 0, SEEK_END);
       size = ftell (files [file]);
       fseek (files [file], pos, SEEK_SET);
       return size;
}

int
seekFile (int file, int offset)
{
        if (fseek (files [file], offset, SEEK_SET) == 0)
          return ftell (files [file]);
        return -1;
}

int
tellFile (int file)
{
        return ftell (files [file]);
}

int
endOfFile (int file)
{
        int c = fgetc (files [file]);
        if (c == EOF)
          return 1;
        ungetc (c, files [file]);
        return 0;
}

extern unsigned long long int currentDispatchID;

int
saveImageNamed (struct ByteArray * name)
{
        char nameString [FILE_NAME_LENGTH];
        size_t nameLength = PSObject_payloadSize ((struct Object *) name);
        FILE * file;
        ImageHeader header;

        if (nameLength >= sizeof (nameString))
          return -1;

        memcpy (nameString, name -> elements, nameLength);
        nameString [nameLength] = '\0';

        file = fopen (nameString, "wb");
        if (file == NULL)
          return -1;

       PSObjectHeap_compact_ (CurrentMemory, 0);
       
       if (CurrentMemory -> markColor == 0)
         PSObjectHeap_garbageCollect (CurrentMemory);

       PSObjectHeap_sweep (CurrentMemory);
       
       PSObjectHeap_adjustAllOopsBy_ (CurrentMemory, - (unsigned) CurrentMemory -> memory);

       header.magicNumber = MAGIC_NUMBER;
       header.imageSize = CurrentMemory -> lastAllocated - (unsigned) CurrentMemory -> memory;
       header.nextHash = CurrentMemory -> lastHash;
       header.specialOops = CurrentMemory -> specialObjectsOop - (unsigned) CurrentMemory -> memory; 
       header.currentDispatchID = currentDispatchID;

       fwrite (& header, sizeof (header), 1, file);
       fwrite (CurrentMemory -> memory, 1, header.imageSize, file);

       fclose (file);

       exit (EXIT_SUCCESS);

       return -1;
}

